﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Shapes;
using Infragistics.Controls.Charts;
using System.Collections.Generic;
using System.Linq;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Windows.Controls.Primitives;
using System.Windows.Data;
using System.Collections;
using System.Windows.Media;

namespace XamWebDataChartSLApp
{
    public class ChartBehaviors : DependencyObject
    {
        #region Crosshair behavior
        public static readonly DependencyProperty CrosshairHidingProperty =
            DependencyProperty.RegisterAttached("CrosshairHiding",
            typeof(CrosshairHidingBehavior), typeof(ChartBehaviors),
            new PropertyMetadata(null,
                (o, e) => CrosshairHidingChanged(
                    o as XamDataChart,
                    e.OldValue as CrosshairHidingBehavior,
                    e.NewValue as CrosshairHidingBehavior)));

        public static CrosshairHidingBehavior GetCrosshairHiding(DependencyObject target)
        {
            return target.GetValue(CrosshairHidingProperty) as CrosshairHidingBehavior;
        }

        public static void SetCrosshairHiding(DependencyObject target, CrosshairHidingBehavior behavior)
        {
            target.SetValue(CrosshairHidingProperty, behavior);
        }

        private static void CrosshairHidingChanged(
            XamDataChart chart,
            CrosshairHidingBehavior oldValue,
            CrosshairHidingBehavior newValue)
        {
            if (chart == null)
            {
                return;
            }
            if (oldValue != null)
            {
                oldValue.OnDetach(chart);
            }
            if (newValue != null)
            {
                newValue.OnAttach(chart);
            }
        }
        #endregion

        #region Tooltip behavior
        public static readonly DependencyProperty CursorTooltipProperty =
            DependencyProperty.RegisterAttached("CursorTooltip",
            typeof(CursorTooltipBehavior), typeof(ChartBehaviors),
            new PropertyMetadata(null,
                (o, e) => CursorTooltipChanged(
                    o as XamDataChart,
                    e.OldValue as CursorTooltipBehavior,
                    e.NewValue as CursorTooltipBehavior)));

        public static CursorTooltipBehavior GetCursorTooltip(DependencyObject target)
        {
            return target.GetValue(CursorTooltipProperty) as CursorTooltipBehavior;
        }

        public static void SetCursorTooltip(DependencyObject target, CursorTooltipBehavior behavior)
        {
            target.SetValue(CursorTooltipProperty, behavior);
        }

        private static void CursorTooltipChanged(
            XamDataChart chart,
            CursorTooltipBehavior oldValue,
            CursorTooltipBehavior newValue)
        {
            if (chart == null)
            {
                return;
            }
            if (oldValue != null)
            {
                oldValue.OnDetach(chart);
            }
            if (newValue != null)
            {
                newValue.OnAttach(chart);
            }
        }
        #endregion

        #region FloatingToolTip Behavior
        public static readonly DependencyProperty FloatingTooltipProperty =
           DependencyProperty.RegisterAttached("FloatingTooltip",
           typeof(FloatingTooltipBehavior), typeof(ChartBehaviors),
           new PropertyMetadata(null,
               (o, e) => FloatingTooltipChanged(
                   o as XamDataChart,
                   e.OldValue as FloatingTooltipBehavior,
                   e.NewValue as FloatingTooltipBehavior)));

        public static FloatingTooltipBehavior GetFloatingTooltip(DependencyObject target)
        {
            return target.GetValue(FloatingTooltipProperty)
                as FloatingTooltipBehavior;
        }

        public static void SetFloatingTooltip(DependencyObject target, FloatingTooltipBehavior behavior)
        {
            target.SetValue(FloatingTooltipProperty, behavior);
        }

        private static void FloatingTooltipChanged(
            XamDataChart chart,
            FloatingTooltipBehavior oldValue,
            FloatingTooltipBehavior newValue)
        {
            if (chart == null)
            {
                return;
            }
            if (oldValue != null)
            {
                oldValue.OnDetach(chart);
            }
            if (newValue != null)
            {
                newValue.OnAttach(chart);
            }
        }
        #endregion

        #region HoveredCategoryItem Behavior
        public static readonly DependencyProperty HoveredCategoryItemsProperty =
           DependencyProperty.RegisterAttached("HoveredCategoryItems",
           typeof(HoveredCategoryItemsBehavior), typeof(ChartBehaviors),
           new PropertyMetadata(null,
               (o, e) => HoveredCategoryItemsChanged(
                   o as XamDataChart,
                   e.OldValue as HoveredCategoryItemsBehavior,
                   e.NewValue as HoveredCategoryItemsBehavior)));

        public static HoveredCategoryItemsBehavior GetHoveredCategoryItems(DependencyObject target)
        {
            return target.GetValue(HoveredCategoryItemsProperty)
                as HoveredCategoryItemsBehavior;
        }

        public static void SetHoveredCategoryItems(DependencyObject target, HoveredCategoryItemsBehavior behavior)
        {
            target.SetValue(HoveredCategoryItemsProperty, behavior);
        }

        private static void HoveredCategoryItemsChanged(
            XamDataChart chart,
            HoveredCategoryItemsBehavior oldValue,
            HoveredCategoryItemsBehavior newValue)
        {
            if (chart == null)
            {
                return;
            }
            if (oldValue != null)
            {
                oldValue.OnDetach(chart);
            }
            if (newValue != null)
            {
                newValue.OnAttach(chart);
            }
        }
        #endregion
    }

    public class CrosshairHidingBehavior
    {
        private Style _invisibleStyle;
        public CrosshairHidingBehavior()
        {
            _invisibleStyle = new Style(typeof(Line));
            _invisibleStyle.Setters.Add(
                new Setter(Shape.StrokeThicknessProperty, 0));
        }

        public bool HideHorizontalCrossHair { get; set; }
        public bool HideVerticalCrossHair { get; set; }

        public void OnAttach(XamDataChart chart)
        {
            chart.MouseEnter += Chart_MouseEnter;
        }

        public void OnDetach(XamDataChart chart)
        {
            chart.MouseEnter -= Chart_MouseEnter;
            MakeVisible(chart, _vertical);
            MakeVisible(chart, _horizontal);
            _vertical = null;
            _horizontal = null;
            _first = true;
        }

        private bool _first = true;
        private Line _vertical;
        private Line _horizontal;

        private void MakeInvisible(
            XamDataChart chart, Line crosshairLine)
        {
            crosshairLine.Style = _invisibleStyle;
        }
        private void MakeVisible(
            XamDataChart chart, Line crosshairLine)
        {
            if (chart == null || crosshairLine == null)
            {
                return;
            }
            crosshairLine.Style = chart.CrosshairLineStyle;
        }

        void Chart_MouseEnter(object sender, MouseEventArgs e)
        {
            XamDataChart chart = sender as XamDataChart;
            if (_first && chart != null)
            {
                var crosshairs =
                    from line in chart.VisualDescendants()
                    where line is Line &&
                          (line as Line).Style
                          == chart.CrosshairLineStyle
                    select line as Line;
                List<Line> lines = crosshairs.ToList();

                // sometimes all values are 0. In that case try it again until we 
                // have one of the lines with different x values
                if (lines[0].X1 == lines[0].X2 && lines[1].X1 == lines[1].X2)
                    return;

                if (lines[0].X1 == lines[0].X2)
                {
                    _vertical = lines[0];
                    _horizontal = lines[1];
                }
                else
                {
                    _vertical = lines[1];
                    _horizontal = lines[0];
                }

                if (HideHorizontalCrossHair)
                {
                    MakeInvisible(chart, _horizontal);
                }
                if (HideVerticalCrossHair)
                {
                    MakeInvisible(chart, _vertical);
                }
                _first = false;
            }
        }
    }

    public class SeriesItemInfo : INotifyPropertyChanged
    {
        private Series _series;
        public Series Series
        {
            get { return _series; }
            set
            {
                _series = value;
                if (PropertyChanged != null)
                {
                    PropertyChanged(
                        this,
                        new PropertyChangedEventArgs("Series"));
                }
            }
        }

        private object _item;
        public object Item
        {
            get { return _item; }
            set
            {
                _item = value;
                if (PropertyChanged != null)
                {
                    PropertyChanged(
                        this,
                        new PropertyChangedEventArgs("Item"));
                }
            }
        }

        public event PropertyChangedEventHandler PropertyChanged;
    }
    public class SeriesItemInfoCollection : ObservableCollection<SeriesItemInfo>
    {
        public void UpdateSeriesItem(Series series, object item)
        {
            bool found = false;
            var indexes = from curr in this
                          where curr.Series.Title == series.Title
                          select this.IndexOf(curr);
            foreach (int index in indexes)
            {
                found = true;
                this[index].Item = item;
            }

            if (!found)
            {
                this.Add(
                    new SeriesItemInfo()
                    {
                        Series = series,
                        Item = item
                    });
            }
        }
    }

    public class CursorTooltipBehavior
    {
        private bool _isOverChart = false;
        private Popup _popup = new Popup();
        private ContentControl _content = new ContentControl();
        private Panel _container;
        private XamDataChart _owner = null;

        private DataTemplate _tooltipTemplate;

        private SeriesItemInfoCollection _items = new SeriesItemInfoCollection();

        public DataTemplate TooltipTemplate
        {
            get { return _tooltipTemplate; }
            set
            {
                _tooltipTemplate = value;
                _content.ContentTemplate = _tooltipTemplate;
            }
        }

        protected bool IsOverChart
        {
            get { return _isOverChart; }
            set
            {
                bool last = _isOverChart;
                _isOverChart = value;
                if (_isOverChart && !last)
                {
                    ShowPopup();
                }
                if (!_isOverChart && last)
                {
                    HidePopup();
                }
            }
        }

        private void HidePopup()
        {
            _popup.IsOpen = false;
        }

        private void ShowPopup()
        {
            _popup.IsOpen = true;
        }

        public void OnAttach(XamDataChart chart)
        {
            if (_owner != null)
            {
                OnDetach(_owner);
            }
            chart.MouseLeave += Chart_MouseLeave;
            chart.MouseMove += Chart_MouseMove;
            chart.SeriesCursorMouseMove += Chart_SeriesCursorMouseMove; ;

            _popup.IsOpen = false;
            _popup.Child = _content;
            _content.ContentTemplate = TooltipTemplate;
            _content.Content = _items;
            if (chart.Parent != null && chart.Parent is Panel)
            {
                _container = chart.Parent as Panel;
                _container.Children.Add(_popup);
            }
        }

        public void OnDetach(XamDataChart chart)
        {
            if (_owner != chart)
            {
                return;
            }
            chart.MouseLeave -= Chart_MouseLeave;
            chart.MouseMove -= Chart_MouseMove;
            chart.SeriesCursorMouseMove -= Chart_SeriesCursorMouseMove;
            IsOverChart = false;
            _items.Clear();
            _owner = null;
        }

        void Chart_MouseMove(object sender, MouseEventArgs e)
        {
            SetPopupOffsets(e.GetPosition(_owner));
            IsOverChart = true;
        }

        private void SetPopupOffsets(Point point)
        {
            _popup.VerticalOffset = point.Y;
            _popup.HorizontalOffset = point.X + 10;
        }

        void Chart_SeriesCursorMouseMove(object sender, ChartCursorEventArgs e)
        {
            if (e.Series != null && e.Item != null)
            {
                _items.UpdateSeriesItem(e.Series, e.Item);
            }
        }

        void Chart_MouseLeave(object sender, MouseEventArgs e)
        {
            IsOverChart = false;
            _items.Clear();
        }
    }

    public class HoveredCategoryItemsBehavior : FrameworkElement
    {
        public HoveredCategoryItemsBehavior()
        {
            HoveredItems = new SeriesItemInfoCollection();
        }

        private XamDataChart _owner = null;

        public static readonly DependencyProperty HoveredItemsProperty =
            DependencyProperty.Register("HoveredItems",
            typeof(SeriesItemInfoCollection),
            typeof(HoveredCategoryItemsBehavior),
            new PropertyMetadata(null));

        public SeriesItemInfoCollection HoveredItems
        {
            get { return (SeriesItemInfoCollection)GetValue(HoveredItemsProperty); }
            set { SetValue(HoveredItemsProperty, value); }
        }

        public bool CategoryOffset { get; set; }

        public void OnAttach(XamDataChart chart)
        {
            if (_owner != null)
            {
                OnDetach(_owner);
            }
            _owner = chart;
            chart.PropertyUpdated += Chart_PropertyUpdated;
        }

        public void OnDetach(XamDataChart chart)
        {
            if (_owner != chart)
            {
                return;
            }
            chart.PropertyUpdated -= Chart_PropertyUpdated;
            HoveredItems.Clear();
            _owner = null;
        }

        private void GetInViewAxisBounds(
            CategoryAxisBase x, NumericYAxis y,
            Rect viewportRect,
            out double left, out double right,
            out double top, out double bottom)
        {
            left = x.GetUnscaledValue(
                viewportRect.Left,
                _owner.WindowRect,
                viewportRect);
            right = x.GetUnscaledValue(
                viewportRect.Right,
                _owner.WindowRect,
                viewportRect);
            top = y.GetUnscaledValue(
                viewportRect.Top,
                _owner.WindowRect,
                viewportRect);
            bottom = y.GetUnscaledValue(
                viewportRect.Bottom,
                _owner.WindowRect,
                viewportRect);
        }

        private Rect GetViewportRect(
            Series target,
            CategoryAxisBase x,
            NumericYAxis y)
        {
            double top = y.TransformToVisual(target)
                    .Transform(new Point(0, 0)).Y;
            double bottom = y.TransformToVisual(target)
                .Transform(new Point(0, y.ActualHeight)).Y;
            double left = x.TransformToVisual(target)
                .Transform(new Point(0, 0)).X;
            double right = x.TransformToVisual(target)
                .Transform(new Point(x.ActualWidth, 0)).X;

            double width = right - left;
            double height = bottom - top;

            if (width > 0.0 && height > 0.0)
            {
                return new Rect(left, top, width, height);
            }
            return Rect.Empty;
        }

        private object GetItem(int index, Series series)
        {
            object item;
            if (series.ItemsSource is IList && index < ((IList)series.ItemsSource).Count)
            {
                item = ((IList)series.ItemsSource)[index];
            }
            else
            {
                IEnumerator items = series.ItemsSource.GetEnumerator();
                for (int i = 0; i < index+1; i++)
                {
                    items.MoveNext();
                }
                item = items.Current;
            }
            return item;
        }

        void Chart_PropertyUpdated(object sender, PropertyUpdatedEventArgs e)
        {
            if (e.PropertyName == "CrosshairPoint")
            {
                foreach (CategorySeries series in
                    _owner.Series.Where((s) => s is CategorySeries))
                {
                    CategoryAxisBase x = series.XAxis;
                    NumericYAxis y = series.YAxis;
                    Rect viewportRect = GetViewportRect(
                       series,
                       x,
                       y);
                    Point p = (Point)e.NewValue;

                    double left;
                    double right;
                    double top;
                    double bottom;
                    GetInViewAxisBounds(x, y, viewportRect,
                        out left, out right,
                        out top, out bottom);
                    double windowX = (p.X - _owner.WindowRect.Left) / _owner.WindowRect.Width;
                    double xAxisValue = left + (windowX * (right - left));
                    if (CategoryOffset)
                    {
                        xAxisValue -= .5;
                    }
                    xAxisValue = Math.Round(xAxisValue);
                    object item = null;

                    if (!double.IsInfinity(xAxisValue) &&
                        !double.IsNaN(xAxisValue) &&
                        xAxisValue >= 0 &&
                        series.ItemsSource != null)
                    {
                        int index = (int)xAxisValue;
                        item = GetItem(index, series);

                        if (item != null)
                        {
                            HoveredItems.UpdateSeriesItem(series, item);
                        }
                    }
                }
            }
        }
    }

    public class FloatingTooltipBehavior : FrameworkElement
    {
        public static readonly DependencyProperty HoveredItemsProperty =
            DependencyProperty.Register("HoveredItems",
            typeof(SeriesItemInfoCollection),
            typeof(FloatingTooltipBehavior),
            new PropertyMetadata(null));

        public SeriesItemInfoCollection HoveredItems
        {
            get { return (SeriesItemInfoCollection)GetValue(HoveredItemsProperty); }
            set { SetValue(HoveredItemsProperty, value); }
        }

        private bool _isOverChart = false;
        private Popup _popup = new Popup();
        private ContentControl _content = new ContentControl();
        private Panel _container;
        private XamDataChart _owner = null;

        private DataTemplate _tooltipTemplate;

        public DataTemplate TooltipTemplate
        {
            get { return _tooltipTemplate; }
            set
            {
                _tooltipTemplate = value;
                _content.ContentTemplate = _tooltipTemplate;
            }
        }

        protected bool IsOverChart
        {
            get { return _isOverChart; }
            set
            {
                bool last = _isOverChart;
                _isOverChart = value;
                if (_isOverChart && !last)
                {
                    ShowPopup();
                }
                if (!_isOverChart && last)
                {
                    HidePopup();
                }
            }
        }

        private void HidePopup()
        {
            _popup.IsOpen = false;
        }

        private void ShowPopup()
        {
            _popup.IsOpen = true;
        }

        public void OnAttach(XamDataChart chart)
        {
            if (_owner != null)
            {
                OnDetach(_owner);
            }
            _owner = chart;
            this.DataContext = chart;
            chart.MouseLeave += Chart_MouseLeave;
            chart.MouseMove += Chart_MouseMove;

            _popup.IsOpen = false;
            _popup.Child = _content;
            _content.ContentTemplate = TooltipTemplate;
            if (_content.Content == null)
            {
                _content.SetBinding(
                    ContentControl.ContentProperty,
                    new Binding("HoveredItems")
                    {
                        Source = this
                    });
            }

            if (chart.Parent != null && chart.Parent is Panel)
            {
                _container = chart.Parent as Panel;
                _container.Children.Add(_popup);
            }
        }

        public void OnDetach(XamDataChart chart)
        {
            if (_owner != chart)
            {
                return;
            }
            chart.MouseLeave -= Chart_MouseLeave;
            chart.MouseMove -= Chart_MouseMove;
            IsOverChart = false;
            if (_container != null)
            {
                _container.Children.Remove(_popup);
            }
            _owner = null;
        }

        void Chart_MouseMove(object sender, MouseEventArgs e)
        {
            // if you pass the chart control in the point is relative to the chart
            // rather than the root control
            //SetPopupOffsets(e.GetPosition(_owner));
            SetPopupOffsets(e.GetPosition(null));
            IsOverChart = true;
        }

        private void SetPopupOffsets(Point point)
        {
            _popup.VerticalOffset = point.Y;

            // need to move the popup from the right side of the pointer
            // to the left side so it does not slide out of view. Problem
            // is that we do not know the size of the popup at this time.
            if (point.X + 30 >= _owner.ActualWidth)
            {
                _popup.HorizontalOffset = point.X - 170;
            }
            else
                _popup.HorizontalOffset = point.X + 10;
        }

        void Chart_MouseLeave(object sender, MouseEventArgs e)
        {
            IsOverChart = false;
            HoveredItems.Clear();
        }
    }

    public static class Extensions
    {
        public static IEnumerable<DependencyObject> VisualDescendants(
            this DependencyObject current)
        {
            yield return current;

            for (int i = 0;
                i < VisualTreeHelper.GetChildrenCount(current);
                i++)
            {
                DependencyObject child =
                    VisualTreeHelper.GetChild(current, i);
                foreach (
                    DependencyObject sub in child.VisualDescendants())
                {
                    yield return sub;
                }
            }
        }
    }
}
